#!/usr/bin/env python3
#
#  textures.py
"""
Texture extraction and conversion logic.
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from collections.abc import Callable
from enum import Enum
from typing import TYPE_CHECKING, cast

# 3rd party
import texture2ddecoder
from PIL import Image

# this package
from cp2077_extractor.cr2w.datatypes import rendRenderTextureBlobPC
from cp2077_extractor.cr2w.enums import ETextureCompression

if TYPE_CHECKING:
	# this package
	from cp2077_extractor.cr2w.datatypes import CBitmapTexture

__all__ = ["DDSFormat", "get_dds_decoder", "get_dds_format_from_compression", "texture_to_image"]


class DDSFormat(Enum):
	"""
	Enum of different formats for DirectDraw Surface ``.dds`` files.
	"""

	R8G8B8A8_UNORM = 0
	BC1_UNORM = 1
	BC3_UNORM = 2
	BC7_UNORM = 7
	BC4_UNORM = 4
	BC5_UNORM = 5


def get_dds_decoder(dds_format: DDSFormat) -> Callable[[bytes, int, int], bytes]:
	"""
	Returns the function to decode a DDS texture, for use with :meth:`PIL.Image.frombytes`.

	:param dds_format:
	"""

	if dds_format == DDSFormat.R8G8B8A8_UNORM:
		raise NotImplementedError
	elif dds_format == DDSFormat.BC1_UNORM:
		return texture2ddecoder.decode_bc1
	elif dds_format == DDSFormat.BC3_UNORM:
		return texture2ddecoder.decode_bc3
	elif dds_format == DDSFormat.BC7_UNORM:
		return texture2ddecoder.decode_bc7
	elif dds_format == DDSFormat.BC4_UNORM:
		return texture2ddecoder.decode_bc4
	elif dds_format == DDSFormat.BC5_UNORM:
		return texture2ddecoder.decode_bc5
	else:
		raise ValueError(f"Unknown format {dds_format!r}")


def get_dds_format_from_compression(compression: ETextureCompression) -> DDSFormat:
	"""
	Find the DDS format for the given texture compression type.

	:param compression
	"""

	if compression == ETextureCompression.TCM_None:
		return DDSFormat.R8G8B8A8_UNORM

	if compression == ETextureCompression.TCM_DXTNoAlpha:
		return DDSFormat.BC1_UNORM
	if compression == ETextureCompression.TCM_Normals:
		return DDSFormat.BC1_UNORM

	if compression == ETextureCompression.TCM_DXTAlpha:
		return DDSFormat.BC3_UNORM
	if compression == ETextureCompression.TCM_NormalsHigh:
		return DDSFormat.BC3_UNORM
	if compression == ETextureCompression.TCM_NormalsGloss:
		return DDSFormat.BC3_UNORM

	if compression == ETextureCompression.TCM_QualityColor:
		return DDSFormat.BC7_UNORM

	if compression == ETextureCompression.TCM_QualityR:
		return DDSFormat.BC4_UNORM

	if compression == ETextureCompression.TCM_QualityRG:
		return DDSFormat.BC5_UNORM
	if compression == ETextureCompression.TCM_Normalmap:
		return DDSFormat.BC5_UNORM

	# if compression == ETextureCompression.TCM_DXTAlphaLinear:
	# 	pass
	# if compression == ETextureCompression.TCM_RGBE:
	# 	pass

	else:
		raise NotImplementedError(compression)


def texture_to_image(texture: "CBitmapTexture") -> Image.Image:
	"""
	Convert a texture to a PIL image.

	:param texture:
	"""

	texture_data = cast(
			rendRenderTextureBlobPC,
			texture.render_texture_resource.render_resource_blob_pc["data"],
			).texture_data["bytes"]
	size = (texture.width, texture.height)
	decoder = get_dds_decoder(get_dds_format_from_compression(texture.setup.compression))

	decoded_data = decoder(texture_data, *size)

	# TODO: check params against other formats
	img = Image.frombytes("RGBA", size, decoded_data, "raw", ("BGRA")).transpose(Image.Transpose.FLIP_TOP_BOTTOM)
	return img



#!/usr/bin/env python3
#
#  anim.py
"""
Classes to represent datatypes within CR2W/W2RC files (prefixed ``anim``).
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

# this package
from cp2077_extractor.cr2w import enums
from cp2077_extractor.cr2w.datatypes.base import Chunk, QsTransform, Quaternion

if TYPE_CHECKING:
	# this package
	from cp2077_extractor.cr2w.datatypes.physics import physicsRagdollBodyInfo, physicsRagdollBodyNames

__all__ = [
		"animAdditionalFloatTrackContainer",
		"animAdditionalFloatTrackEntry",
		"animAdditionalTransformContainer",
		"animAdditionalTransformEntry",
		"animAnimDataChunk",
		"animAnimEvent",
		"animAnimFallbackFrameDesc",
		"animAnimFeature",
		"animAnimSet",
		"animAnimSetEntry",
		"animAnimSetup",
		"animAnimSetupEntry",
		"animAnimation",
		"animBoneCorrection",
		"animCompareBone",
		"animEventsContainer",
		"animFacialEmotionTransitionBaked",
		"animFloatTrackInfo",
		"animIAnimationBuffer",
		"animIKTargetRequest",
		"animIMotionExtraction",
		"animIRigIkSetup",
		"animLookAtLimits",
		"animLookAtPartRequest",
		"animLookAtRequest",
		"animLookAtRequestForPart",
		"animPoseCorrection",
		"animPoseCorrectionGroup",
		"animRig",
		"animRigPart",
		"animRigPartBone",
		"animRigPartBoneTree",
		"animRigRetarget",
		"animTransformInfo",
		"animTransformMask",
		]


class animIAnimationBuffer(Chunk):
	pass


@dataclass
class animTransformInfo(Chunk):
	name: str = ''
	parent_name: str = ''
	reference_transform_ls: QsTransform = field(
			default_factory=lambda: QsTransform(translation=(0.0, 0.0, 0.0, 1.0))
			)


@dataclass
class animAdditionalTransformEntry(Chunk):
	transform_info: animTransformInfo = field(default_factory=animTransformInfo)
	value: QsTransform = field(default_factory=QsTransform)


@dataclass
class animAdditionalTransformContainer(Chunk):
	entries: list[animAdditionalTransformEntry] = field(default_factory=list)


@dataclass
class animFloatTrackInfo(Chunk):
	name: str = ''
	reference_value: float = 0.0


@dataclass
class animAdditionalFloatTrackEntry(Chunk):
	name: str = ''
	track_info: animFloatTrackInfo = field(default_factory=animFloatTrackInfo)
	values: list[float] = field(default_factory=list)


@dataclass
class animAdditionalFloatTrackContainer(Chunk):
	entries: list[animAdditionalFloatTrackEntry] = field(default_factory=list)
	overwrite_existing_values: bool = True


class animIMotionExtraction(Chunk):
	pass


@dataclass
class animAnimation(Chunk):
	tags: list = field(default_factory=list)
	name: str = ''
	duration: float = 0.0
	animation_type: enums.animAnimationType = enums.animAnimationType.Normal
	anim_buffer: animIAnimationBuffer = field(default_factory=animIAnimationBuffer)
	additional_transforms: animAdditionalTransformContainer = field(
			default_factory=animAdditionalTransformContainer
			)
	additional_tracks: animAdditionalFloatTrackContainer = field(default_factory=animAdditionalFloatTrackContainer)
	motion_extraction: animIMotionExtraction = field(default_factory=animIMotionExtraction)
	frame_clamping: bool = False
	frame_clamping_start_frame: int = -1
	frame_clamping_end_frame: int = -1


@dataclass
class animAnimEvent(Chunk):
	start_frame: int = 0
	duration_in_frames: int = 0
	event_name: str = ''


@dataclass
class animEventsContainer(Chunk):
	events: list[animAnimEvent] = field(default_factory=list)


@dataclass
class animAnimSetEntry(Chunk):
	animation: animAnimation = field(default_factory=animAnimation)
	events: animEventsContainer = field(default_factory=animEventsContainer)


@dataclass
class animAnimDataChunk(Chunk):
	buffer: bytes = b''


@dataclass
class animAnimFallbackFrameDesc(Chunk):
	mpositions: int = 0
	mrotations: int = 0
	mfloat_tracks: int = 0


@dataclass
class animRigPartBone(Chunk):
	bone: str = ''
	weight: float = 1.0


@dataclass
class animRigPartBoneTree(Chunk):
	root_bone: str = ''
	weight: float = 1.0
	subtrees_to_change: list["animRigPartBoneTree"] = field(default_factory=list)


@dataclass
class animTransformMask(Chunk):
	index: int = -1
	weight: float = 0.0


@dataclass
class animRigPart(Chunk):
	name: str = ''
	single_bones: list[animRigPartBone] = field(default_factory=list)
	tree_bones: list[animRigPartBoneTree] = field(default_factory=list)
	bones_with_rotation_in_model_space: list[str] = field(default_factory=list)
	mask: list[animTransformMask] = field(default_factory=list)
	mask_rot_ms: list[int] = field(default_factory=list)


@dataclass
class animIRigIkSetup(Chunk):
	name: str = ''


@dataclass
class animIKTargetRequest(Chunk):
	weight_position: float = 1.0
	weight_orientation: float = 1.0
	transition_in: float = 0.3
	transition_out: float = 0.3
	priority: int = 0


@dataclass
class animRig(Chunk):
	cooking_platform: enums.ECookingPlatform = enums.ECookingPlatform.PLATFORM_None
	bone_names: list[str] = field(default_factory=list)
	track_names: list[str] = field(default_factory=list)
	rig_extra_tracks: list[animFloatTrackInfo] = field(default_factory=list)
	level_of_detail_start_indices: list[int] = field(default_factory=list)
	distance_category_to_lod_map: list[int] = field(default_factory=list)
	turn_off_lod: int = -1
	turning_off_update_and_sample: bool = False
	reference_tracks: list[float] = field(default_factory=list)
	reference_pose_ms: list[QsTransform] = field(default_factory=list)
	apose_ls: list[QsTransform] = field(default_factory=list)
	apose_ms: list[QsTransform] = field(default_factory=list)
	tags: list = field(default_factory=list)
	parts: list[animRigPart] = field(default_factory=list)
	retargets: list["animRigRetarget"] = field(default_factory=list)
	ik_setups: list[animIRigIkSetup] = field(default_factory=list)
	ragdoll_desc: list["physicsRagdollBodyInfo"] = field(default_factory=list)
	ragdoll_names: list["physicsRagdollBodyNames"] = field(default_factory=list)


@dataclass
class animRigRetarget(Chunk):
	source_rig: animRig = field(default_factory=animRig)  # TODO: CResourceReference


@dataclass
class animAnimSet(Chunk):
	cooking_platform: enums.ECookingPlatform = enums.ECookingPlatform.PLATFORM_None
	animations: list[animAnimSetEntry] = field(default_factory=list)
	animation_data_chunks: list[animAnimDataChunk] = field(default_factory=list)
	fallback_data_addresses: list[int] = field(default_factory=list)
	fallback_data_address_indexes: list[int] = field(default_factory=list)
	fallback_anim_frame_descs: list[animAnimFallbackFrameDesc] = field(default_factory=list)
	fallback_anim_desc_indexes: list[int] = field(default_factory=list)
	fallback_anim_data_buffer: Any = None  # TODO: DataBuffer = field(default_factory=DataBuffer)
	fallback_num_position_data: int = 0
	fallback_num_rotation_data: int = 0
	fallback_num_float_track_data: int = 0
	rig: animRig = field(default_factory=animRig)  # TODO: CResourceReference
	tags: list = field(default_factory=list)
	version: int = 0


@dataclass
class animFacialEmotionTransitionBaked(Chunk):
	to_idle_male: str = ''
	facial_key_male: str = ''
	to_idle_female: str = ''
	facial_key_female: str = ''
	transition_type: enums.animFacialEmotionTransitionType = enums.animFacialEmotionTransitionType.Fast
	transition_duration: float = 0.0
	time_scale: float = 0.0
	to_idle_weight: float = 1.0
	to_idle_neck_weight: float = 0.0
	facial_key_weight: float = 0.0
	custom_transition_anim: str = ''


@dataclass
class animLookAtLimits(Chunk):
	soft_limit_degrees: float = 360.0
	hard_limit_degrees: float = 360.0
	hard_limit_distance: float = 1000000.0
	back_limit_degrees: float = 180.0


class animAnimFeature(Chunk):
	pass


@dataclass
class animLookAtPartRequest(Chunk):
	part_name: str = ''
	weight: float = 0.5
	suppress: float = 0.0
	mode: int = 0


@dataclass
class animLookAtRequest(Chunk):
	transition_speed: float = 60.0
	has_out_transition: bool = False
	out_transition_speed: float = 60.0
	following_speed_factor_override: float = -1.0
	limits: animLookAtLimits = field(default_factory=animLookAtLimits)
	suppress: float = 0.0
	mode: int = 0
	calculate_position_in_parent_space: bool = False
	priority: int = 0
	additional_parts: animLookAtPartRequest = field(default_factory=animLookAtPartRequest)
	invalid: bool = False


@dataclass
class animLookAtRequestForPart(Chunk):
	body_part: str = "Eyes"
	request: animLookAtRequest = field(default_factory=animLookAtRequest)
	attach_left_hand_to_right_hand: int = -1
	attach_right_hand_to_left_hand: int = -1


@dataclass
class animCompareBone(Chunk):
	bone_name: str = ''
	bone_rotation_ls: Quaternion = field(default_factory=Quaternion)


@dataclass
class animBoneCorrection(Chunk):
	bone_name: str = ''
	additive_correction: Quaternion = field(default_factory=Quaternion)


@dataclass
class animPoseCorrection(Chunk):
	rbf_coefficient: float = 3.5
	rbf_pow_value: float = 20.0
	compare_bones: animCompareBone = field(default_factory=animCompareBone)
	bone_corrections: animBoneCorrection = field(default_factory=animBoneCorrection)


@dataclass
class animPoseCorrectionGroup(Chunk):
	pose_corrections: animPoseCorrection = field(default_factory=animPoseCorrection)


@dataclass
class animAnimSetupEntry(Chunk):
	priority: int = 128
	anim_set: animAnimSet = field(default_factory=animAnimSet)  # TODO: CResourceAsyncReference
	variable_names: list[str] = field(default_factory=list)


@dataclass
class animAnimSetup(Chunk):
	cinematics: list[animAnimSetupEntry] = field(default_factory=list)
	gameplay: list[animAnimSetupEntry] = field(default_factory=list)
	hash: int = 0



#!/usr/bin/env python3
#
#  rend.py
"""
Classes to represent datatypes within CR2W/W2RC files (prefixed ``rend``).
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from dataclasses import dataclass, field
from typing import Any

# this package
from cp2077_extractor.cr2w import enums
from cp2077_extractor.cr2w.datatypes.base import Chunk, DeferredBufferData, HandleData

__all__ = [
		"rendRenderTextureBlobHeader",
		"rendRenderTextureBlobMemoryLayout",
		"rendRenderTextureBlobMipMapInfo",
		"rendRenderTextureBlobPC",
		"rendRenderTextureBlobPlacement",
		"rendRenderTextureBlobSizeInfo",
		"rendRenderTextureBlobTextureInfo",
		"rendRenderTextureResource",
		"rendSLightFlickering",
		]


@dataclass
class rendRenderTextureBlobTextureInfo(Chunk):
	texture_data_size: int
	slice_size: int
	data_alignment: int
	slice_count: int
	mip_count: int
	type: enums.GpuWrapApieTextureType = enums.GpuWrapApieTextureType.TEXTYPE_2D


@dataclass
class rendRenderTextureBlobSizeInfo(Chunk):
	"""
	Size info for a texture.
	"""

	width: int
	height: int
	depth: int = 1


@dataclass
class rendRenderTextureBlobHeader(Chunk):
	"""
	Header for texture data and associated properties.
	"""

	version: int
	size_info: rendRenderTextureBlobSizeInfo
	texture_info: rendRenderTextureBlobTextureInfo
	flags: int
	mip_map_info: list[Any] = field(default_factory=list)  # list[MipMapInfo]  # TODO: parse array
	histogram_data: list[Any] = field(default_factory=list)  # list[HistogramData]


@dataclass
class rendRenderTextureBlobPC(Chunk):
	header: rendRenderTextureBlobHeader
	texture_data: DeferredBufferData


@dataclass
class rendRenderTextureResource(Chunk):
	render_resource_blob_pc: HandleData[Chunk] = field(
			default_factory=dict
			)  # type: ignore[assignment] # TODO: handle inner type # CHandle


@dataclass
class rendRenderTextureBlobMemoryLayout(Chunk):
	row_pitch: int = 0
	slice_pitch: int = 0


@dataclass
class rendRenderTextureBlobPlacement(Chunk):
	size: int = 0
	offset: int = 0


@dataclass
class rendRenderTextureBlobMipMapInfo(Chunk):
	layout: rendRenderTextureBlobMemoryLayout = field(default_factory=rendRenderTextureBlobMemoryLayout)
	placement: rendRenderTextureBlobPlacement = field(default_factory=rendRenderTextureBlobPlacement)


@dataclass
class rendSLightFlickering(Chunk):
	position_offset: float = 0.0
	flicker_strength: float = 0.0
	flicker_period: float = 0.2



#!/usr/bin/env python3
#
#  utils.py
"""
Utility functions.
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import struct
from io import BytesIO
from typing import Any

# this package
from cp2077_extractor.cr2w.header_structs import CR2WFileInfo

__all__ = ["get_array_variables", "get_chunk_variables", "get_names_list"]


def get_names_list(file_info: CR2WFileInfo) -> list[bytes]:
	"""
	Returns the name lookup table for the file.

	:param file_info:
	"""

	_names_list: list[bytes] = []
	for a_name_info in file_info.name_info:
		assert a_name_info.offset in file_info.string_dict
		_names_list.append(file_info.string_dict[a_name_info.offset])

	return _names_list


def get_chunk_variables(chunk: bytes, names_list: list[bytes]) -> list[tuple[bytes, bytes, Any]]:
	"""
	Parse variables from the given chunk.

	:param chunk:
	:param names_list: Name lookup table for the file.

	:returns: List of variables, as tuples of variable name, variable REDengine type, variable value.
	"""

	return _read_class(BytesIO(chunk), len(chunk), names_list)


def get_array_variables(chunk: bytes, names_list: list[bytes]) -> list[list[tuple[bytes, bytes, Any]]]:
	"""
	Parse variables for an array from the given chunk.

	:param chunk:
	:param names_list: Name lookup table for the file.

	:returns: List of lists of variables, as tuples of variable name, variable REDengine type, variable value.
	"""

	buffer = BytesIO(chunk)
	array_size = int.from_bytes(buffer.read(4), "little")
	variables: list[list[tuple[bytes, bytes, Any]]] = []

	for _ in range(array_size):
		# variables.append(_read_class(buffer, len(chunk), names_list))
		variables.append(_read_class(buffer, 0, names_list))

	assert buffer.tell() == len(chunk)

	return variables


def _read_class(buffer: BytesIO, chunk_size: int, names_list: list[bytes]) -> list[tuple[bytes, bytes, Any]]:
	# this package
	from cp2077_extractor.cr2w.io import CNameError, read_c_name

	zero = buffer.read(1)[0]
	assert zero == 0, f"Tried parsing a CVariable: zero read {zero}."

	variables: list[tuple[bytes, bytes, Any]] = []
	# while buffer.tell() < chunk_size:
	while True:
		try:
			var_c_name = read_c_name(buffer, names_list)
		except CNameError:
			break
		red_type_name = read_c_name(buffer, names_list)
		size = struct.unpack("<I", buffer.read(4))[0] - 4
		value = buffer.read(size)
		variables.append((var_c_name, red_type_name, value))

	if chunk_size:
		assert buffer.tell() == chunk_size, (buffer.tell(), chunk_size)

	return variables



#!/usr/bin/env python3
#
#  mesh.py
"""
Classes to represent datatypes within CR2W/W2RC files (prefixed ``mesh``).
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from dataclasses import dataclass, field
from typing import Any

# this package
from cp2077_extractor.cr2w import enums
from cp2077_extractor.cr2w.datatypes.base import Box, Chunk, CMatrix, HandleData

__all__ = [
		"CMaterialInstance",
		"CMesh",
		"CMeshMaterialEntry",
		"IMaterial",
		"meshLocalMaterialHeader",
		"meshMeshAppearance",
		"meshMeshMaterialBuffer",
		"meshMeshParameter",
		]


@dataclass
class IMaterial(Chunk):
	cooking_platform: enums.ECookingPlatform = enums.ECookingPlatform.PLATFORM_None


@dataclass
class CMaterialInstance(IMaterial):
	base_material: IMaterial = field(default_factory=IMaterial)  # TODO: CResourceReference
	enable_mask: bool = False
	audio_tag: str = ''
	resource_version: int = 0


@dataclass
class meshMeshAppearance(Chunk):
	name: str = "default"
	chunk_materials: list[str] = field(default_factory=list)
	tags: list[str] = field(default_factory=list)


@dataclass
class meshLocalMaterialHeader(Chunk):
	offset: int = 0
	size: int = 0


@dataclass
class meshMeshMaterialBuffer(Chunk):
	raw_data: Any = None  # TODO: DataBuffer = field(default_factory=DataBuffer)
	raw_data_headers: list[meshLocalMaterialHeader] = field(default_factory=list)


class meshMeshParameter(Chunk):
	pass


@dataclass
class CMeshMaterialEntry(Chunk):
	name: str = ''
	index: int = 0
	is_local_instance: bool = False


@dataclass
class CMesh(Chunk):
	cooking_platform: enums.ECookingPlatform = enums.ECookingPlatform.PLATFORM_None
	# Missing ordinal 1
	parameters: list[meshMeshParameter] = field(default_factory=list)
	bounding_box: Box = field(
			default_factory=Box
			)  # TODO: Box { Min = new Vector4 { X = float.MaxValue, Y = float.MaxValue, Z = float.MaxValue, W = float.MaxValue }, Max = new Vector4 { X = float.MinValue, Y = float.MinValue, Z = float.MinValue, W = float.MinValue } };
	surface_area_per_axis: tuple[float, float, float] = (-1.0, -1.0, -1.0)
	# Missing ordinal 5
	material_entries: list[CMeshMaterialEntry] = field(default_factory=list)
	external_materials: list[IMaterial] = field(default_factory=list)  # TODO: CResourceAsyncReference
	local_material_instances: list[CMaterialInstance] = field(default_factory=list)
	local_material_buffer: meshMeshMaterialBuffer = field(default_factory=meshMeshMaterialBuffer)
	preload_external_materials: list[IMaterial] = field(default_factory=list)  # TODO: CResourceReference
	preload_local_material_instances: list[IMaterial] = field(default_factory=list)
	inplace_resources: list[Chunk] = field(default_factory=list)  # TODO: CResourceReference
	appearances: list[meshMeshAppearance] = field(default_factory=list)
	object_type: enums.ERenderObjectType = enums.ERenderObjectType.ROT_Static
	render_resource_blob: HandleData[Chunk] = field(  # TODO: Handle inner type
		default_factory=dict
		)  # type: ignore[assignment]  # IRenderResourceBlob
	lod_level_info: list[float] = field(default_factory=list)
	float_track_names: list[str] = field(default_factory=list)
	bone_names: list[str] = field(default_factory=list)
	bone_rig_matrices: list[CMatrix] = field(default_factory=list)
	bone_vertex_epsilons: list[float] = field(default_factory=list)
	lod_bone_mask: list[int] = field(default_factory=list)
	# Missing ordinal 22
	constrain_auto_hide_distance_to_terrain_height_map: bool = False
	force_load_all_appearances: bool = False
	cast_global_shadows_cached_in_cook: bool = False
	cast_local_shadows_cached_in_cook: bool = False
	use_ray_tracing_shadow_lodbias: bool = False
	casts_ray_traced_shadows_from_original_geometry: bool = False
	is_shadow_mesh: bool = False
	is_player_shadow_mesh: bool = False



#!/usr/bin/env python3
#
#  appearance.py
"""
Classes to represent datatypes within CR2W/W2RC files (prefixed ``appearance``).
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import sys
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

# this package
from cp2077_extractor.cr2w import enums
from cp2077_extractor.cr2w.datatypes.base import Chunk, Transform, redTagList
from cp2077_extractor.cr2w.datatypes.mesh import CMesh

if TYPE_CHECKING:
	# this package
	from cp2077_extractor.cr2w.datatypes.ent import (
			entdismembermentEffectResource,
			entdismembermentWoundResource,
			entdismembermentWoundsConfigSet,
			entEntityParametersBuffer,
			entEntityTemplate
			)
	from cp2077_extractor.cr2w.datatypes.game import gameHitRepresentationOverride

__all__ = [
		"appearanceAlternateAppearanceEntry",
		"appearanceAppearanceDefinition",
		"appearanceAppearancePart",
		"appearanceAppearancePartOverrides",
		"appearanceAppearanceResource",
		"appearanceCensorshipEntry",
		"appearanceCookedAppearanceData",
		"appearancePartComponentOverrides",
		]


@dataclass
class appearanceAlternateAppearanceEntry(Chunk):
	original: str = ''
	alternate: str = ''
	alternate_appearance_index: int = 0


@dataclass
class appearanceCensorshipEntry(Chunk):
	original: str = ''
	censored: str = ''
	censor_flags: int = 0


@dataclass
class appearancePartComponentOverrides(Chunk):
	component_name: str = ''
	mesh_appearance: str = "default"
	chunk_mask: int = sys.maxsize
	use_custom_transform: bool = False
	initial_transform: Transform = field(default_factory=Transform)
	visual_scale: tuple[float, float, float] = (0.0, 0.0, 0.0)
	accept_dismemberment: bool = True


@dataclass
class appearanceCookedAppearanceData(Chunk):
	cooking_platform: enums.ECookingPlatform = enums.ECookingPlatform.PLATFORM_None
	dependencies: list[Chunk] = field(default_factory=list)  # TODO: CResourceReference
	total_size_on_disk: int = 0


@dataclass
class appearanceAppearancePart(Chunk):
	resource: "entEntityTemplate" = field(
			default_factory=lambda: entEntityTemplate()
			)  # TODO: resolve default class (circular import)  # TODO: CResourceAsyncReference


@dataclass
class appearanceAppearancePartOverrides(Chunk):
	part_resource: "entEntityTemplate" = field(
			default_factory=lambda: entEntityTemplate()
			)  # TODO: resolve default class (circular import)    # TODO: CResourceAsyncReference
	components_overrides: list[appearancePartComponentOverrides] = field(default_factory=list)


@dataclass
class appearanceAppearanceDefinition(Chunk):
	name: str = ''
	parent_appearance: str = ''
	parts_masks: list[list[str]] = field(default_factory=list)
	parts_values: list[appearanceAppearancePart] = field(default_factory=list)
	parts_overrides: list[appearanceAppearancePartOverrides] = field(default_factory=list)
	proxy_mesh: CMesh = field(default_factory=CMesh)  # TODO: CResourceAsyncReference
	forced_lod_distance: int = 0
	proxy_mesh_appearance: str = ''
	cooked_data_path_override: Chunk = field(default_factory=Chunk)  # TODO: CResourceAsyncReference
	parameters_buffer: "entEntityParametersBuffer" = field(
			default_factory=lambda: entEntityParametersBuffer()
			)  # TODO: resolve default class (circular import)
	visual_tags: redTagList = field(default_factory=redTagList)
	inherited_visual_tags: redTagList = field(default_factory=redTagList)
	hit_representation_overrides: list["gameHitRepresentationOverride"] = field(default_factory=list)
	compiled_data: Any = None  # TODO: SerializationDeferredDataBuffer = field(default_factory=SerializationDeferredDataBuffer)
	resolved_dependencies: list[Chunk] = field(default_factory=list)  # TODO: CResourceAsyncReference
	loose_dependencies: list[Chunk] = field(default_factory=list)  # TODO: CResourceAsyncReference
	censor_flags: int = 0


class appearanceAppearanceResource(Chunk):
	cooking_platform: enums.ECookingPlatform = enums.ECookingPlatform.PLATFORM_None
	alternate_appearance_setting_name: str = ''
	alternate_appearance_suffixes: list[str] = field(default_factory=list)
	alternate_appearance_mapping: list[appearanceAlternateAppearanceEntry] = field(default_factory=list)
	censorship_mapping: list[appearanceCensorshipEntry] = field(default_factory=list)
	wounds: list["entdismembermentWoundResource"] = field(default_factory=list)
	dism_effects: list["entdismembermentEffectResource"] = field(default_factory=list)
	dism_wound_config: "entdismembermentWoundsConfigSet" = field(
			default_factory=lambda: entdismembermentWoundsConfigSet()
			)  # TODO: resolve default class (circular import)
	base_type: str = ''
	base_entity_type: str = ''
	base_entity: "entEntityTemplate" = field(
			default_factory=lambda: entEntityTemplate()
			)  # TODO: resolve default class (circular import)   # TODO: CResourceAsyncReference
	part_type: str = ''
	preset: str = ''
	appearances: list[appearanceAppearanceDefinition] = field(default_factory=list)
	common_cook_data: appearanceCookedAppearanceData = field(
			default_factory=appearanceCookedAppearanceData
			)  # TODO: CResourceAsyncReference
	proxy_poly_count: int = 1400
	force_compile_proxy: bool = False
	generate_player_blocking_collision_for_proxy: bool = False



#!/usr/bin/env python3
#
#  advert_data.py
"""
Data for audio for adverts that play on animated billboards.
"""
#
#  Copyright © 2025 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import NamedTuple

__all__ = ["AdvertData", "adverts"]


class AdvertData(NamedTuple):
	"""
	Data for an advert's sound files.
	"""

	#: The filenames in the game files (in ``base/localization/en-us/vo`` or ``base\localization\common\vo``) for this advert all start with this string.
	audio_filename_prefix: str

	#: The name of the scene file in ``base/media/animated_billboards/scenes`` in the game files (with the suffix ``.scene``.)
	scene_file: str

	general_audio: bool = False
	"""
	Whether the audio is common too all game languages.

	:py:obj:`False` for localised audio, :py:obj:`True` for files in ``audio_1_general.archive``.
	"""


#: Data for audio for adverts that play on animated billboards.
adverts: dict[str, AdvertData] = {
		"all_foods": AdvertData("ad_all_foods_ab_ad_caliente", "ab_q003_01_all_foods_meat_ad"),
		"arasaka": AdvertData(
				"ad_arasaka_ab_ad_caliente",
				"ab_sq032_01_arasaka_propaganda",
				),  # ab_q115_01_arasaka_propaganda
		"captain_caliente": AdvertData("ad_captain_caliente_ab_ad_caliente", "ab_ad_caliente"),  # TODO: Check
		"chromanticore": AdvertData("ad_chromanticore_ab_ad_chromanticore", "ab_ad_chromanticore"),
		"foreign_body": AdvertData("ad_foreign_body_ab_ad_foreign_body", "ab_ad_foreign_body"),
		"kang_tao": AdvertData("ad_kang_tao_ab_ad_caliente", "ab_q104_01_kang_tao_ad"),
		"lizzies_bar": AdvertData("ad_lizzies_bar_ab_ad_caliente", "ab_q004_01_lizzies_bar_ad"),
		"mrstud": AdvertData("ad_mrstud_ab_ad_mrstud", "ab_ad_mrstud"),
		"mrwhitey": AdvertData("ad_mrwhitey_ab_ad_mrwhitey", "ab_ad_mrwhitey"),
		"nicola": AdvertData("ad_nicola_female_ab_ad_nicola", "ab_ad_nicola"),
		"night_corp": AdvertData("ad_night_corp_ab_ad_caliente", "ab_q114_01_night_corp_ad"),
		"orgiatic": AdvertData("ad_orgiatic_male_ab_ad_orgiatic", "ab_ad_orgiatic"),
		"slaughterhouse": AdvertData("ad_slaughterhouse_ab_ad_slaughterhouse", "ab_ad_slaughterhouse"),
		"sojasil": AdvertData("ad_sojasil_ab_ad_sojasil", "ab_ad_sojasil"),
		"televangelist": AdvertData("ad_televangelist_ab_ad_caliente", "ab_sts_wat_nid_04_televangelist_ad"),
		"thrud": AdvertData("ad_thrud_ab_ad_thrud", "ab_ad_thrud"),
		"tiancha": AdvertData("ad_tiancha_ab_ad_tiancha", "ab_ad_tiancha"),
		"vargas": AdvertData("ad_vargas_ab_ad_vargas", "ab_ad_vargas"),
		"watson_whore": AdvertData("ad_watson_whore_ab_ad_caliente", "ab_ad_watson_whore"),
		"us_cracks": AdvertData("civ_mid_f_60_jap_25_ab_ad_caliente", "ab_sq017_01_us_cracks_ad"),
		"crystal_palace": AdvertData("ziggy_q_q203_01_crystal_palace_info", "q203_01_crystal_palace_info"),
		"delamain": AdvertData("delamain_ab_ad_caliente", "ab_sq025_01_delamain_ad"),  #
		# TODO: 'jefferson_peralez': AdvertData('jefferson_peralez_jefferson_peralez_ad', ),
		}
